/**
 * NORRIS-IoC.js
 * Copyright (c) 2011 Ruben LZ Tan
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
 * IN THE SOFTWARE.
 */
'use strict';

///////////////////////////////////////////////////////////////////////////////
// @nordoc[package] ioc
//
// version
// -    refactored code to use the standard NORRIS creational pattern
// -    hoisted all variables
// -    changed the event 'error' to 'failure'
// -    added documentation
///////////////////////////////////////////////////////////////////////////////
module.exports.make = function make(deps) {
    if (undefined === deps) deps = {};

    ///////////////////////////////////////////////////////////////////////////
    // Hoisted variables
    ///////////////////////////////////////////////////////////////////////////
    var base, npm, fs, Emitter, Installer, IoC, _loadNpm, _loadNode;

    ///////////////////////////////////////////////////////////////////////////
    // Dependency injection/definition
    ///////////////////////////////////////////////////////////////////////////
    base        = deps.base         || process.cwd(),
    npm         = deps.npm          || require('npm'),
    fs          = deps.fs           || require('fs'),
    Emitter     = deps.Emitter      || require('events').EventEmitter,
    Installer   = deps.Installer    || require(__dirname + '/installer');

    //-------------------------------------------------------------------------
    // @nordoc[function]
    //
    // @desc    Loads an NPM module, installs it if necessary
    //
    // @args    key [string]
    //          The key in the dependency object where the loaded npm module
    //          will be saved to
    //
    //          module [string]
    //          The module to load
    //
    //          cb [function] (err, key, returnedModule)
    //          The callback to call upon completion of the operation
    //
    // @return  none
    //-------------------------------------------------------------------------
    _loadNpm = function _loadNpm(key, module, cb) {
        var self = this;

        self.installer.install(module, function(err, npmMod) {
            if (err) {
                cb(err);
                return;
            }

            cb(null, key, npmMod);
            return;
        });
    }

    //-------------------------------------------------------------------------
    // @nordoc[function]
    //
    // @desc    Loads a core node module
    //
    // @args    key [string]
    //          The key in the dependency object where the loaded node module
    //          will be saved to
    //
    //          module [string]
    //          The module to load
    //
    //          cb [function] (err, key, returnedModule)
    //          The callback to call upon completion of the operation
    //
    // @return  none
    //-------------------------------------------------------------------------
    _loadNode = function _loadNode(key, module, cb) {
        var nodeMod;

        try {
            nodeMod = require(module);
            cb(null, key, nodeMod);
        } catch(err) {
            cb(err);
        }
    }

    ///////////////////////////////////////////////////////////////////////////
    // @nordoc[module] ioc
    ///////////////////////////////////////////////////////////////////////////
    IoC = {
        //---------------------------------------------------------------------
        // @nordoc[property]
        //
        // @desc    The base path/current working directory of the application
        //---------------------------------------------------------------------
        basepath : base,

        //---------------------------------------------------------------------
        // @nordoc[property]
        //
        // @desc    The NPM installer
        //---------------------------------------------------------------------
        installer : Installer.make({
            npm : npm,
            fs  : fs
        }),

        //---------------------------------------------------------------------
        // @nordoc[property]
        //
        // @desc    Container of loader functions for the various registered
        //          signatures
        //---------------------------------------------------------------------
        loader : {
            npm     : _loadNpm, 
            node    : _loadNode
        },

        //---------------------------------------------------------------------
        // @nordoc[property]
        //
        // @desc    Dependencies currently loaded
        //---------------------------------------------------------------------
        deps : {},

        //---------------------------------------------------------------------
        // @nordoc[property]
        //
        // @desc    Logs
        //---------------------------------------------------------------------
        log : [],

        //---------------------------------------------------------------------
        // @nordoc[method]
        //
        // @desc    Registers a new signature
        //
        // @args    signature [string]
        //          The name of the signature to be registered
        //
        //          cb [function] (key, module, callback)
        //          The callback to be called when encountering this specific
        //          load signature
        //
        // @return  Returns itself so that other methods can be chained
        //---------------------------------------------------------------------
        register : function register(signature, cb) {
            this.loader[signature]= cb;
            return this;
        },

        //---------------------------------------------------------------------
        // @nordoc[method]
        //
        // @desc    Adds a dependency object to load loading list
        //
        // @args    deps [object]
        //          A dependency object
        //
        // @return  Returns itself so that other methods can be chained
        //---------------------------------------------------------------------
        load : function load(deps) {
            var loadList, log, key, dep;

            loadList    = this.deps;
            log         = this.log;

            // Mixin to the current load list
            for (key in deps) {
                dep = deps[key];

                // Log any duplicates in the loading list
                if (undefined !== loadList[key])
                    log.push('Duplicate load list: ' + key + ' -> ' + dep);
                    
                loadList[key] = dep;
            }

            this.deps   = loadList;
            this.log    = log;

            return this;
        },

        //---------------------------------------------------------------------
        // @nordoc[method]
        //
        // @desc    Starts the loading process. Upon completion, it will call
        //          the callback passed in as an argument.
        //
        // @args    cb [function] (deps)
        //
        //
        // @return  none
        //---------------------------------------------------------------------
        ready : function ready(cb) {
            var deps, self, result, promise, types, loaded, npmConfig;

            deps    = this.deps;
            self    = this;
            result  = new Emitter();
            promise = new Emitter();
            types   = Object.keys(this.loader);
            loaded  = {};

            npmConfig = {
                loglevel    : 'silent',
                outfd       : ''
            }

            if ('' !== this.basepath) {
                npmConfig.prefix = this.basepath;
                self.installer.basepath = this.basepath;
            }

            npm.load(npmConfig, function onMake() {
                var key, dep, start, end, type, module;

                // If callbacks are not used, pump in a dummy function
                if (undefined === cb) cb = function() {};

                // Propagate the error event to the resulting callback
                promise.on('failure', function onError(err) {
                    process.nextTick(function() {
                        result.emit('error', err, module);
                        cb(err);
                    });
                });

                // Add the successful loading of the module to the returned dep object
                promise.on('success', function onSuccess(key, module) {
                    loaded[key] = module;
                    delete deps[key];

                    if (0 === Object.keys(deps).length) {
                        process.nextTick(function() {
                            result.emit('success', loaded);
                            cb(null, loaded);
                        });
                    }
                });

                // Run through all the dependencies, loading as we go
                for (key in deps) {
                    dep = deps[key];

                    // Modules are always split into two parts with a colon ':'
                    start   = dep.indexOf('[');
                    end     = dep.indexOf(']');
                    type    = dep.slice(start + 1, end);
                    module  = dep.slice(end + 1).trim();

                    if (!type || !module) {
                        continue;
                    }

                    // Use the proper loader function
                    if (-1 !== types.indexOf(type)) {
                        self.loader[type].call(self, key, module, function onLoaded(err, key, mod) {
                            if (err) {
                                promise.emit('failure', err);
                            } else {
                                promise.emit('success', key, mod);
                            }
                        });
                    }
                }    
            });

            return result;
        }
    }

    return IoC;
}